/**ElevenDB.java
 * Abstracts away how the server handles its database operations.  Currently
 * uses a Vector implementation.  May later use a Hashtable.*/

import java.util.*;
import java.io.*;

public class ElevenDB {
  private Hashtable theDB;  
  public final int DEFAULT_INIT_SIZE = 200; //expect no more than 200 users
  public final int MAX_TRANSACTIONS_REMEMBERED = 100;
  public final int MAX_MESSAGES_REMEMBERED = 100;
  public static final String DELIM = ":";

  /**Create a default, empty database.  Not all that useful
   * since creating accounts is not implemented*/
  public ElevenDB() {
    theDB = new Hashtable(DEFAULT_INIT_SIZE);
  }

  /**Creates a database built from a file.  The file format is very simple:
   * a name, a tab, and an integer value, optional nick and pass if true is 
   * passed as an argument.  History is blank to start with.
   * # at the beginning of a line represents a comment.
   * If false, then the file is read through a ObjectInputStream.*/
  public ElevenDB(String filename, boolean isText) {
    theDB = new Hashtable(DEFAULT_INIT_SIZE);
    try {
      if (isText) {
	BufferedReader buff = new BufferedReader(new FileReader (filename));
	String input = buff.readLine();
	while (input != null) {
	  if (input.length() == 0 || input.charAt(0) == '#') {
	    input = buff.readLine();
	    continue; //comments and blanks skipped
	  }
	  String[] parsedData = input.split(DELIM);  //colon
	  /*
	    System.err.print("Parsed :");
	    for(int i=0; i<parsedData.length; i++) System.err.print(parsedData[i] + ":");
	    System.err.println();
	  */
	  String name = parsedData[1]; //first value is key, so skipped
	  int value = Integer.parseInt(parsedData[2]); //second val
	  String nick = "DEFAULT_NICK";
	  String pass = "DEFAULT_PASS";
	  String quote = "NO_QUOTE";
	  String trades = "NO_TRADES";
	  String email = "NO_EMAIL";
	  String screen = "NO_SCREEN";
	  if (parsedData.length > 3) {
	    nick = parsedData[3];
	    pass = parsedData[4];
	  }
	  if (parsedData.length > 5) {
	    email = parsedData[5];
	    trades = parsedData[6];
	  }
	  System.err.println("Adding user:" + name + " " + value + " " + nick + " " + pass + " " + quote + " " + trades + " "+ email + " " + screen);
	  theDB.put(name, new ElevenDBData(name, value, nick, pass, quote, trades, email, screen));
	  input = buff.readLine();
	}
	buff.close();
      } else {  //ObjectInputStream
	//from Java API

	FileInputStream fis = new FileInputStream(filename);
	ObjectInputStream ois = new ObjectInputStream(fis);
	theDB = (Hashtable) ois.readObject();
	ois.close();

      }
    } catch (FileNotFoundException e) {
      System.err.println(e.getMessage());
      System.err.println("Using blank DB");
    } catch (IOException e) {
      System.err.println(e.getMessage());
      System.err.println("Input ceased--input file is possibly corrupt.\n" +  
			 "DB possibly incomplete.");
    } catch (ClassNotFoundException e) {
      System.err.println(e.getMessage());
      System.err.println("Class does not match!  Using blank DB");
    }
  }
  /*Write out the DB into a file.*/
  public void writeOut(String filename, boolean inText) {
    try {
      if (inText) {
	PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
	Enumeration k = theDB.keys();
	Enumeration e = theDB.elements();
	while (e.hasMoreElements()) {
	  out.print((String) k.nextElement());
	  out.print(DELIM);
	  out.println((ElevenDBData) e.nextElement());
	}
	out.flush();
      } else {
	//Java API
	FileOutputStream fos = new FileOutputStream(filename);
	ObjectOutputStream oos = new ObjectOutputStream(fos);
	oos.writeObject(theDB);
	oos.close();
      }
    } catch (IOException e) {
      System.err.println("writeOut to file " + filename + " fails--continuing!");
      e.printStackTrace();
    }
  }

  /**Modifies the value of a user by a certain increment.  If the user
   * is invalid, the operation is ignored and -1 is returned.*/
  public int modifyValue(String user, int amountToAdd) {
    int current = getValue(user);
    if (current != -1) {
      setValue(user, current + amountToAdd);
      return 0;
    }

    return -1;
  }

  /*Return the associated value of the username.  Returns -1 if fails.*/
  public int getValue(String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      return entry.getValue();
    }
    return -1;
  }

  public void setValue(String user, int value) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      entry.setValue(value);
    }
  }

  /**Selects between getNamePass and getDetails*/
  public String getInfo(String user, boolean complete, boolean namePass) {
    if (namePass) 
      return getNamePass(user, complete);
    else 
      return getDetails(user, complete);
  }

  /**Returns a string of data.  If complete is true, returns all fields.  If false,
   * returns some of the fields hidden.*/
  public String getNamePass(String user, boolean complete) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      if (complete) {
	return entry.netStringInfo();
      } else return entry.netStringInfoHidden();
    }
    return "";
  }

  /**Returns a string of data.  If complete is true, returns all fields.  If false,
   * returns some of the fields hidden.*/
  public String getDetails(String user, boolean complete) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      if (complete) {
	return entry.netStringDetails();
      } else return entry.netStringDetailsHidden();
    }
    return "";
  }

  /**Set the nick and pwd of a given user.  Returns true if user exists, 
   * false otherwise.*/
  public boolean setInfo(String user, String nick, String pwd) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      entry.setNick(nick);
      entry.setPwd(pwd);
      return true;
    }
    return false;
  }


  /**Modifies the details of a person's info.*/
  public boolean setDetails(String user, String quote, String trades, 
                            String email, String screen) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      entry.setQuote(quote);
      entry.setTrades(trades);
      entry.setEmail(email);
      entry.setScreen(screen);
      return true;
    }
    return false;
  }

  /**Sets the lastLogin to be the current time.*/
  public boolean updateLastLogin(String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    if (entry != null) {
      entry.updateLastLogin();
      return true;
    }
    return false;    
  }


  /**Adds an entry to a user's history, given a user and the transaction.  Removes the
   * oldest entry if it's bigger than the constant, so that its size remains limited.*/
  public void addToHistory(String ownerOfHistory, String user, String targetUser, 
			   int amount, String memo) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(ownerOfHistory));
    if (entry != null) {
      entry.getHistory().addFirst(new ElevenDBTransaction(user, targetUser, amount, memo));
      if (entry.getHistory().size() > MAX_TRANSACTIONS_REMEMBERED) {
	entry.getHistory().removeLast();
      }
    }
  }

  /**Another addToHistory method, but this time the date is available.*/
  public void addToHistory(String ownerOfHistory, String user, String targetUser, 
			   int amount, String memo, String date) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(ownerOfHistory));
    if (entry != null) {
      entry.getHistory().addFirst(new ElevenDBTransaction(user, targetUser, amount, memo, date));
      if (entry.getHistory().size() > MAX_TRANSACTIONS_REMEMBERED) {
	entry.getHistory().removeLast();
      }
    }
  }

  /**Add a buddy to a buddy list.*/
  public void addToBuddyList(String ownerOfList, String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(ownerOfList));
    if (!searchBuddies(ownerOfList, user)) {
      entry.getBuddyList().addFirst(user);
    }
  }
  /**See if a username is in a given buddy list.*/
  public boolean searchBuddies(String ownerOfList, String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(ownerOfList));
    ListIterator temp = entry.getBuddyList().listIterator();
    while ((entry != null) && (temp.hasNext())) {
      if (((String) temp.next()).equals(user)) return true;
    }
    return false;
  }

  /**Add a message to someone's list.  The buddy should already be in the user's
     Buddy list--this is checked in ElevenServer.*/
  public void addMessage(String recipient, String from, String message) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(recipient));
    if (entry != null) {
      entry.getMessages().addLast(from + "(" + Calendar.getInstance().getTime() + ")" +
                                  message);
      if (entry.getMessages().size() > MAX_MESSAGES_REMEMBERED) {
        entry.getMessages().removeLast();
      }
    }
  }

  /**Get the last logins of all your buddies.*/
  public String getBuddyInfos(String username) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(username));
    String buddyInfo = "";
    if (entry != null) {
      ListIterator temp = entry.getBuddyList().listIterator();
      while ((entry != null) && (temp.hasNext())) {
        String buddy = (String) temp.next();
        ElevenDBData buddyData = (ElevenDBData) theDB.get(buddy);
        if (buddyData.getLastLogin() == null) {
          buddyInfo = buddy + " has never logged in.\0" + buddyInfo;
        } else {
          buddyInfo = buddy + " last logged in on " + buddyData.getLastLogin().toString()
            + "\0" + buddyInfo; //HACKHACK see ElevenClient
        }
      }
    }
    return buddyInfo;
  }

  /**Get the messages left for a user, concatenate them together with \0s and return
     them.*/
  public String getMessages(String username) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(username));
    String messages = "";
    if (entry != null) {
      ListIterator temp = entry.getMessages().listIterator();
      while ((entry != null) && (temp.hasNext())) {
        messages = (String) temp.next() + "\0" + messages;
      }
    }
    return messages;
  }


  /**Resets the History Iterator so it returns values from the beginning again.*/
  /*
    public void resetHistoryIterator(String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    System.err.println(entry.getHistory().size());
    if (entry != null) {
    entry.setIterator(entry.getHistory().listIterator());
    }
    }*/

  /**Iterator reader--returns null if no more elements.*/
  public String getHistory(String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    String ans = "";
    ListIterator temp = entry.getIterator();
    while ((entry != null) && (temp.hasNext())) {  //HACKHACK
      ans =  (ElevenDBTransaction) temp.next() + "\0" + ans;//HACKHACK see ElevenClient
    }                                         //can't send newline or it
    return ans;                               //won't read the whole msg!
  }

  /**Returns a user's password.  Used by ElevenServer.  Returns
   * true if the user exists and the password matches, false
   * otherwise.*/
  public boolean checkPassword(String user, String pass) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    return ((entry != null) &&
	    (entry.getPwd().equals(pass)));
  }

  /**Returns a user.  Used for debugging*/
  private ElevenDBData getData(String user) {
    ElevenDBData entry = ((ElevenDBData) theDB.get(user));
    return entry;	    
  }

  /**Test this part.  Create a DB off of a test input, write out, and
   * modify values to writeout again*/
  public static void main(String args[]) {
    ElevenDB newDB = new ElevenDB("SampleDB.txt", true);
    newDB.addToHistory("name1", "name1", "name2", 50, "for fun");
    newDB.modifyValue("name1", 100);
    newDB.modifyValue("name1", 100);
    newDB.modifyValue("name2", -600);
    newDB.modifyValue("name2", -600);
    newDB.modifyValue("cs11faa", -600214);
    newDB.modifyValue("cs11fsasdajiodjaoijdoajodijaojdosajda", -600);
    newDB.writeOut("output.txt", true);
    newDB.writeOut("output2.txt", false);
    ElevenDB secondDB = new ElevenDB("output2.txt", false);
    System.err.println(secondDB.getData("name1").stringHistory());
    secondDB.writeOut("output3.txt", true);
  }
}
